"""Implementation of PaccMann^RL conditional generators."""

import json
import logging
import os
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
import torch
from rdkit import Chem
from paccmann_chemistry.models import StackGRUDecoder, StackGRUEncoder, TeacherVAE
from paccmann_chemistry.utils.search import SamplingSearch
from paccmann_omics.encoders import ENCODER_FACTORY
from pytoda.smiles.smiles_language import SMILESLanguage

from ....domains.materials import validate_molecules
from ....domains.materials.protein_encoding import PrimarySequenceEncoder
from ....frameworks.torch import device_claim
from ....frameworks.torch.vae import reparameterize

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ConditionalGenerator(ABC):
    """Abstract interface for a conditional generator."""

    #: device where the inference is running.
    device: torch.device
    #: temperature for the sampling.
    temperature: float
    #: maximum length of the generated molecules.
    generated_length: int

    #: parameters for the SELFIES generator.
    selfies_conditional_generator_params: dict
    #: SELFIES generator.
    selfies_conditional_generator: TeacherVAE
    #: SMILES language instance.
    smiles_language: SMILESLanguage

    generator_latent_size: int
    encoder_latent_size: int

    def get_smiles_from_latent(self, latent: torch.Tensor) -> List[str]:
        """Take samples from the latent space.

        Args:
            latent: latent vector tensor.

        Returns:
            SMILES list and indexes for the valid ones.
        """
        if self.generator_latent_size == 2 * self.encoder_latent_size:
            latent = latent.repeat(1, 1, 2)

        # generate molecules as tokens list
        generated_molecules = self.selfies_conditional_generator.generate(
            latent,
            prime_input=torch.tensor(
                [self.smiles_language.start_index], device=self.device
            ).long(),
            end_token=torch.tensor(
                [self.smiles_language.stop_index], device=self.device
            ).long(),
            generate_len=self.generated_length,
            search=SamplingSearch(temperature=self.temperature),
        )

        # decode SELFIES
        selfies = [
            self.smiles_language.token_indexes_to_smiles(generated_molecule.tolist())
            for generated_molecule in iter(generated_molecules)
        ]

        # convert SELFIES to SMILES
        smiles = [
            self.smiles_language.selfies_to_smiles(a_selfies) for a_selfies in selfies
        ]
        return smiles

    @staticmethod
    def validate_molecules(smiles) -> Tuple[List[Chem.rdchem.Mol], List[int]]:
        return validate_molecules(smiles_list=smiles)

    @abstractmethod
    def get_latent(self, condition: Any) -> torch.Tensor:
        pass

    @abstractmethod
    def generate_valid(self, condition: Any, number_of_molecules: int) -> List[str]:
        """
        Generate a given number of samples (molecules) from a given condition.

        Args:
            protein: the protein used as context/condition.
            number_of_molecules: number of molecules to sample.

        Returns:
            list of SMILES generated.
        """
        # prepare the molecule set
        generated_molecules: Set[str] = set()
        logger.info("embedding condition and getting reparametrized latent samples")
        latent = self.get_latent(condition)
        logger.info("starting generation of molecules")
        while len(generated_molecules) < number_of_molecules:
            # generate the molecules
            generated_smiles = self.get_smiles_from_latent(latent)
            _, valid_ids = self.validate_molecules(generated_smiles)
            generated_molecules |= set([generated_smiles[index] for index in valid_ids])
        logger.info("completed generation of molecules")
        # return the molecules listed by length
        return sorted(list(generated_molecules), key=len, reverse=True)[
            :number_of_molecules
        ]

    def generate_batch(self, condition: Any) -> List[str]:
        logger.info("embedding condition and getting reparametrized latent samples")
        latent = self.get_latent(condition)
        logger.info("starting generation of molecules")
        # generate the molecules
        return self.get_smiles_from_latent(latent)


class ProteinSequenceConditionalGenerator(ConditionalGenerator):
    """
    Protein conditional generator as implemented in https://doi.org/10.1088/2632-2153/abe808
    (originally https://arxiv.org/abs/2005.13285).
    It generates highly binding and low toxic ligands.

    Attributes:
        samples_per_protein: number of points sampled per protein.
            It has to be greater than 1.
        protein_embedding_encoder_params: parameter for the protein embedding encoder.
        protein_embedding_encoder: protein embedding encoder.
    """

    def __init__(
        self,
        resources_path: str,
        temperature: float = 1.4,
        generated_length: int = 100,
        samples_per_protein: int = 100,
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Initialize the generator.

        Args:
            resources_path: directory where to find models and parameters.
            temperature: temperature for the sampling. Defaults to 1.4.
            generated_length: maximum length of the generated molecules.
                Defaults to 100.
            samples_per_protein: number of points sampled per protein.
                It has to be greater than 1. Defaults to 10.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        # device
        self.device = device_claim(device)
        # setting sampling parameters
        self.temperature = temperature
        self.generated_length = generated_length
        self.samples_per_protein = samples_per_protein
        # instantiate protein embedding encoder
        with open(os.path.join(resources_path, "protein_embedding_params.json")) as fp:
            self.protein_embedding_encoder_params = json.load(fp)
        self.protein_embedding_encoder = ENCODER_FACTORY["dense"](
            self.protein_embedding_encoder_params
        ).to(self.device)
        self.protein_embedding_encoder.load(
            os.path.join(resources_path, "protein_embedding_encoder.pt"),
            map_location=self.device,
        )
        self.protein_embedding_encoder.eval()
        self.encoder_latent_size = self.protein_embedding_encoder.latent_size
        # instantiate selfies conditional generator
        with open(
            os.path.join(resources_path, "selfies_conditional_generator.json")
        ) as fp:
            self.selfies_conditional_generator_params = json.load(fp)
        self.selfies_conditional_generator = TeacherVAE(
            StackGRUEncoder(self.selfies_conditional_generator_params),
            StackGRUDecoder(self.selfies_conditional_generator_params),
        ).to(self.device)
        self.selfies_conditional_generator.load(
            os.path.join(resources_path, "selfies_conditional_generator.pt"),
            map_location=self.device,
        )
        self.selfies_conditional_generator.eval()
        self.generator_latent_size = (
            self.selfies_conditional_generator.decoder.latent_dim
        )
        # loading SMILES language for decoding and conversion of SELFIES to SMILES
        self.smiles_language = SMILESLanguage.load(
            os.path.join(resources_path, "selfies_language.pkl")
        )
        # protein embedding from primary sequence (via tape)
        self.primary_sequence_embedder = PrimarySequenceEncoder(
            model_type="transformer",
            from_pretrained="bert-base",
            model_config_file=None,
            tokenizer="iupac",
        ).to(self.device)

    def get_latent(self, protein: str) -> torch.Tensor:
        """
        Given a protein generate the latent representation.

        Args:
            protein: the protein used as context/condition.

        Returns:
            the latent representation for the given context. It contains
                self.samples_per_protein repeats.
        """
        # encode embedded sequence once, ignore the returned dummy ids
        embeddings, _ = self.primary_sequence_embedder.forward([[protein]])
        protein_mu, protein_logvar = self.protein_embedding_encoder(
            embeddings.to(self.device)
        )

        # now stack as batch to generate different samples
        proteins_mu = torch.cat([protein_mu] * self.samples_per_protein, dim=0)
        proteins_logvar = torch.cat([protein_logvar] * self.samples_per_protein, dim=0)
        # get latent representation
        return torch.unsqueeze(reparameterize(proteins_mu, proteins_logvar), 0)

    def generate_valid(self, protein: str, number_of_molecules: int) -> List[str]:
        """
        Generate a given number of samples (molecules) from a given protein.

        Args:
            protein: the protein used as context/condition.
            number_of_molecules: number of molecules to sample.

        Returns:
            list of SMILES generated.
        """
        return super().generate_valid(
            condition=protein, number_of_molecules=number_of_molecules
        )

    def generate_batch(self, protein: str) -> List[str]:
        return super().generate_batch(condition=protein)


class TranscriptomicConditionalGenerator(ConditionalGenerator):
    """
    Transcriptomic conditional generator as implemented in https://doi.org/10.1016/j.isci.2021.102269
    (originally https://doi.org/10.1007/978-3-030-45257-5_18, https://arxiv.org/abs/1909.05114).
    It generates highly effective small molecules against transcriptomic progiles.

    Attributes:
        samples_per_profile: number of points sampled per profile.
            It has to be greater than 1.
        transcriptomic_encoder_params: parameter for the protein embedding encoder.
        transcriptomic_encoder: protein embedding encoder.
    """

    def __init__(
        self,
        resources_path: str,
        temperature: float = 1.4,
        generated_length: int = 100,
        samples_per_profile: int = 100,
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Initialize the generator.

        Args:
            resources_path: directory where to find models and parameters.
            temperature: temperature for the sampling. Defaults to 1.4.
            generated_length: maximum length of the generated molecules.
                Defaults to 100.
            samples_per_profile: number of points sampled per protein.
                It has to be greater than 1. Defaults to 10.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        # device
        self.device = device_claim(device)
        # setting sampling parameters
        self.temperature = temperature
        self.generated_length = generated_length
        self.samples_per_profile = samples_per_profile
        with open(os.path.join(resources_path, "genes.txt")) as fp:
            self.genes = [gene.strip() for gene in fp if gene]
        # instantiate protein embedding encoder
        with open(os.path.join(resources_path, "transcriptomic_params.json")) as fp:
            self.transcriptomic_encoder_params = json.load(fp)
        self.transcriptomic_encoder = ENCODER_FACTORY["dense"](
            self.transcriptomic_encoder_params
        ).to(self.device)
        self.transcriptomic_encoder.load(
            os.path.join(resources_path, "transcriptomic_encoder.pt"),
            map_location=self.device,
        )
        self.transcriptomic_encoder.eval()
        self.encoder_latent_size = self.transcriptomic_encoder.latent_size
        # instantiate selfies conditional generator
        with open(
            os.path.join(resources_path, "selfies_conditional_generator.json")
        ) as fp:
            self.selfies_conditional_generator_params = json.load(fp)
        self.selfies_conditional_generator = TeacherVAE(
            StackGRUEncoder(self.selfies_conditional_generator_params),
            StackGRUDecoder(self.selfies_conditional_generator_params),
        ).to(self.device)
        self.selfies_conditional_generator.load(
            os.path.join(resources_path, "selfies_conditional_generator.pt"),
            map_location=self.device,
        )
        self.selfies_conditional_generator.eval()
        self.generator_latent_size = (
            self.selfies_conditional_generator.decoder.latent_dim
        )
        # loading SMILES language for decoding and conversion of SELFIES to SMILES
        self.smiles_language = SMILESLanguage.load(
            os.path.join(resources_path, "selfies_language.pkl")
        )

    def get_latent(self, profile: Union[np.ndarray, pd.Series, str]) -> torch.Tensor:
        """
        Given a profile generate the latent representation.

        Args:
            profile: the profile used as context/condition.

        Raises:
            ValueError: in case the profile has a size mismatch with the genes panel.

        Returns:
            the latent representation for the given context. It contains
                self.samples_per_profile repeats.
        """
        if isinstance(profile, pd.Series):
            # make sure genes are sorted
            profile = profile[self.genes].values
        elif isinstance(profile, str):
            logger.warning("profile passed as string, serializing it to a list")
            profile = np.array(json.loads(profile))
        if profile.size != len(self.genes):
            raise ValueError(
                f"provided profile size ({profile.size}) does not match required size {len(self.genes)}"
            )
        # encode embedded progiles
        transcriptomic_mu, transcriptomic_logvar = self.transcriptomic_encoder(
            torch.from_numpy(
                np.vstack([profile] * self.samples_per_profile),
            )
            .float()
            .to(self.device)
        )
        # get latent representation
        return torch.unsqueeze(
            reparameterize(transcriptomic_mu, transcriptomic_logvar), 0
        )

    def generate_valid(
        self, profile: Union[np.ndarray, pd.Series], number_of_molecules: int
    ) -> List[str]:
        """
        Generate a given number of samples (molecules) from a given transcriptomic profile.

        Args:
            profile: the profile used as context/condition.
            number_of_molecules: number of molecules to sample.

        Returns:
            list of SMILES generated.
        """
        return super().generate_valid(
            condition=profile, number_of_molecules=number_of_molecules
        )

    def generate_batch(self, profile: Union[np.ndarray, pd.Series]) -> List[str]:
        return super().generate_batch(condition=profile)



"""Moses Organ implementation."""

import argparse

from guacamol_baselines.moses_baselines.organ_distribution_learning import (
    OrganGenerator,
)


class Organ:
    def __init__(
        self,
        model_path: str,
        model_config_path: str,
        vocab_path: str,
        n_samples: int,
        n_batch: int,
        max_len: int,
    ):
        """Initialize Organ.

        Args:
            model_path: path from where to load the model
            model_config_path: path from where to load the model config
            vocab_path: path from where to load the vocab
            n_samples: Number of samples to sample
            n_batch: Size of the batch
            max_len: Max length of SMILES
        """
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("--model_load", default=model_path)
        self.parser.add_argument("--config_load", default=model_config_path)
        self.parser.add_argument("--vocab_load", default=vocab_path)
        self.parser.add_argument("--n_samples", default=n_samples)
        self.parser.add_argument("--n_batch", default=n_batch)
        self.parser.add_argument("--max_len", default=max_len)
        self.parser.add_argument("--device", default="cpu")
        self.config = self.parser.parse_known_args()[0]

    def get_generator(self) -> OrganGenerator:
        """
        used for creating an instance of the OrganGenerator

        Returns:
            An instance of OrganGenerator
        """
        optimiser = OrganGenerator(self.config)
        return optimiser



"""PaccMann VAE training utilities."""

import json
import logging
import os
from dataclasses import dataclass, field
from time import time
from typing import Any, Dict, Optional, cast

import torch
from paccmann_chemistry.models.training import train_vae
from paccmann_chemistry.models.vae import StackGRUDecoder, StackGRUEncoder, TeacherVAE
from paccmann_chemistry.utils import collate_fn, disable_rdkit_logging
from paccmann_chemistry.utils.hyperparams import SEARCH_FACTORY
from pytoda.datasets import SMILESDataset
from pytoda.smiles.smiles_language import SMILESLanguage
from torch.utils.tensorboard import SummaryWriter

from ....frameworks.torch import get_device
from ...core import TrainingPipelineArguments
from ..core import PaccMannTrainingPipeline

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class PaccMannVAETrainingPipeline(PaccMannTrainingPipeline):
    """Language modeling training pipelines."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for PaccMann training.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.
        """
        try:
            device = get_device()
            disable_rdkit_logging()
            params = {**training_args, **dataset_args, **model_args}
            train_smiles_filepath = params["train_smiles_filepath"]
            test_smiles_filepath = params["test_smiles_filepath"]
            smiles_language_filepath = (
                params["smiles_language_filepath"]
                if params.get("smiles_language_filepath", "none").lower() != "none"
                else None
            )

            model_path = params["model_path"]
            training_name = params["training_name"]

            writer = SummaryWriter(f"logs/{training_name}")
            logger.info(f"Model with name {training_name} starts.")

            model_dir = os.path.join(model_path, training_name)
            log_path = os.path.join(model_dir, "logs")
            val_dir = os.path.join(log_path, "val_logs")
            os.makedirs(os.path.join(model_dir, "weights"), exist_ok=True)
            os.makedirs(os.path.join(model_dir, "results"), exist_ok=True)
            os.makedirs(log_path, exist_ok=True)
            os.makedirs(val_dir, exist_ok=True)

            # Load SMILES language
            smiles_language: Optional[SMILESLanguage] = None
            if smiles_language_filepath is not None:
                smiles_language = SMILESLanguage.load(smiles_language_filepath)

            logger.info(f"Smiles filepath: {train_smiles_filepath}")

            # create SMILES eager dataset
            smiles_train_data = SMILESDataset(
                train_smiles_filepath,
                smiles_language=smiles_language,
                padding=False,
                selfies=params.get("selfies", False),
                add_start_and_stop=params.get("add_start_stop_token", True),
                augment=params.get("augment_smiles", False),
                canonical=params.get("canonical", False),
                kekulize=params.get("kekulize", False),
                all_bonds_explicit=params.get("all_bonds_explicit", False),
                all_hs_explicit=params.get("all_hs_explicit", False),
                remove_bonddir=params.get("remove_bonddir", False),
                remove_chirality=params.get("remove_chirality", False),
                backend="lazy",
                device=device,
            )
            smiles_test_data = SMILESDataset(
                test_smiles_filepath,
                smiles_language=smiles_language,
                padding=False,
                selfies=params.get("selfies", False),
                add_start_and_stop=params.get("add_start_stop_token", True),
                augment=params.get("augment_smiles", False),
                canonical=params.get("canonical", False),
                kekulize=params.get("kekulize", False),
                all_bonds_explicit=params.get("all_bonds_explicit", False),
                all_hs_explicit=params.get("all_hs_explicit", False),
                remove_bonddir=params.get("remove_bonddir", False),
                remove_chirality=params.get("remove_chirality", False),
                backend="lazy",
                device=device,
            )

            if smiles_language_filepath is None:
                smiles_language = smiles_train_data.smiles_language
                smiles_language.save(os.path.join(model_path, f"{training_name}.lang"))
            else:
                smiles_language_filename = os.path.basename(smiles_language_filepath)
                cast(SMILESLanguage, smiles_language).save(
                    os.path.join(model_dir, smiles_language_filename)
                )

            params.update(
                {
                    "vocab_size": cast(
                        SMILESLanguage, smiles_language
                    ).number_of_tokens,
                    "pad_index": cast(SMILESLanguage, smiles_language).padding_index,
                }
            )

            vocab_dict = cast(SMILESLanguage, smiles_language).index_to_token
            params.update(
                {
                    "start_index": list(vocab_dict.keys())[
                        list(vocab_dict.values()).index("<START>")
                    ],
                    "end_index": list(vocab_dict.keys())[
                        list(vocab_dict.values()).index("<STOP>")
                    ],
                }
            )

            if params.get("embedding", "learned") == "one_hot":
                params.update({"embedding_size": params["vocab_size"]})

            with open(os.path.join(model_dir, "model_params.json"), "w") as fp:
                json.dump(params, fp)

            # create DataLoaders
            train_data_loader = torch.utils.data.DataLoader(
                smiles_train_data,
                batch_size=params.get("batch_size", 64),
                collate_fn=collate_fn,
                drop_last=True,
                shuffle=True,
                pin_memory=params.get("pin_memory", True),
                num_workers=params.get("num_workers", 8),
            )

            test_data_loader = torch.utils.data.DataLoader(
                smiles_test_data,
                batch_size=params.get("batch_size", 64),
                collate_fn=collate_fn,
                drop_last=True,
                shuffle=True,
                pin_memory=params.get("pin_memory", True),
                num_workers=params.get("num_workers", 8),
            )
            # initialize encoder and decoder
            gru_encoder = StackGRUEncoder(params).to(device)
            gru_decoder = StackGRUDecoder(params).to(device)
            gru_vae = TeacherVAE(gru_encoder, gru_decoder).to(device)
            logger.info("Model summary:")
            for name, parameter in gru_vae.named_parameters():
                logger.info(f"Param {name}, shape:\t{parameter.shape}")
            total_params = sum(p.numel() for p in gru_vae.parameters())
            logger.info(f"Total # params: {total_params}")

            loss_tracker = {
                "test_loss_a": 10e4,
                "test_rec_a": 10e4,
                "test_kld_a": 10e4,
                "ep_loss": 0,
                "ep_rec": 0,
                "ep_kld": 0,
            }

            # train for n_epoch epochs
            logger.info("Model creation and data processing done, Training starts.")
            decoder_search = SEARCH_FACTORY[params.get("decoder_search", "sampling")](
                temperature=params.get("temperature", 1.0),
                beam_width=params.get("beam_width", 3),
                top_tokens=params.get("top_tokens", 5),
            )

            if writer:
                pparams = params.copy()
                pparams["training_file"] = train_smiles_filepath
                pparams["test_file"] = test_smiles_filepath
                pparams["language_file"] = smiles_language_filepath
                pparams["model_path"] = model_path
                pparams = {k: v if v is not None else "N.A." for k, v in params.items()}
                pparams["training_name"] = training_name
                from pprint import pprint

                pprint(pparams)
                writer.add_hparams(hparam_dict=pparams, metric_dict={})

            for epoch in range(params["epochs"] + 1):
                t = time()
                loss_tracker = train_vae(
                    epoch,
                    gru_vae,
                    train_data_loader,
                    test_data_loader,
                    smiles_language,
                    model_dir,
                    search=decoder_search,
                    optimizer=params.get("optimizer", "adadelta"),
                    lr=params["learning_rate"],
                    kl_growth=params["kl_growth"],
                    input_keep=params["input_keep"],
                    test_input_keep=params["test_input_keep"],
                    generate_len=params["generate_len"],
                    log_interval=params["log_interval"],
                    save_interval=params["save_interval"],
                    eval_interval=params["eval_interval"],
                    loss_tracker=loss_tracker,
                    logger=logger,
                    # writer=writer,
                    batch_mode=params.get("batch_mode"),
                )
                logger.info(f"Epoch {epoch}, took {time() - t:.1f}.")

            logger.info(
                "Overall:\tBest loss = {0:.4f} in Ep {1}, "
                "best Rec = {2:.4f} in Ep {3}, "
                "best KLD = {4:.4f} in Ep {5}".format(
                    loss_tracker["test_loss_a"],
                    loss_tracker["ep_loss"],
                    loss_tracker["test_rec_a"],
                    loss_tracker["ep_rec"],
                    loss_tracker["test_kld_a"],
                    loss_tracker["ep_kld"],
                )
            )
            logger.info("Training done, shutting down.")
        except Exception:
            logger.exception(
                "Exception occurred while running PaccMannVAETrainingPipeline"
            )


@dataclass
class PaccMannVAEModelArguments(TrainingPipelineArguments):
    """Arguments pertaining to model instantiation."""

    __name__ = "model_args"

    n_layers: int = field(
        default=2, metadata={"help": "Number of layers for the RNNs."}
    )
    bidirectional: bool = field(
        default=False, metadata={"help": "Whether the RNN cells are bidirectional."}
    )
    rnn_cell_size: int = field(default=512, metadata={"help": "Size of the RNN cells."})
    latent_dim: int = field(default=256, metadata={"help": "Size of the RNN cells."})
    stack_width: int = field(
        default=50, metadata={"help": "Width of the memory stack for the RNN cell."}
    )
    stack_depth: int = field(
        default=50, metadata={"help": "Depth of the memory stack for the RNN cell."}
    )
    decode_search: str = field(
        default="sampling", metadata={"help": "Decoder search strategy."}
    )
    dropout: float = field(default=0.2, metadata={"help": "Dropout rate to apply."})
    generate_len: int = field(
        default=100, metadata={"help": "Length in tokens of the generated molecules."}
    )
    kl_growth: float = field(
        default=0.003, metadata={"help": "Growth of the KL term weight in the loss."}
    )
    input_keep: float = field(
        default=0.85, metadata={"help": "Probability to keep input tokens in train."}
    )
    test_input_keep: float = field(
        default=1.0, metadata={"help": "Probability to keep input tokens in test."}
    )
    temperature: float = field(
        default=0.8, metadata={"help": "Temperature for the sampling."}
    )
    embedding: str = field(
        default="one_hot",
        metadata={
            "help": "Embedding technique for the tokens. 'one_hot' or 'learned'."
        },
    )
    batch_mode: str = field(
        default="packed", metadata={"help": "Batch mode. 'packed' or 'padded'."}
    )
    vocab_size: int = field(
        default=380, metadata={"help": "Size of the vocabulary of chemical tokens."}
    )
    pad_index: int = field(default=0, metadata={"help": "Index for the padding token."})
    embedding_size: int = field(
        default=380, metadata={"help": "Size of the embedding vectors."}
    )
    beam_width: int = field(default=3, metadata={"help": "Width of the beam search."})
    top_tokens: int = field(
        default=5, metadata={"help": "Number of tokens to consider in the beam search."}
    )



"""Functions to facilitate conversion from dataclasses to training descriptions."""


from dataclasses import _MISSING_TYPE, fields
from typing import Any, Dict, Optional, Type, Union


def find_type(input_type: Type) -> Optional[str]:
    """Convert type class to string.

    Args:
        input_type: Type to be converted to string.

    Returns:
        String of the type or None if the given type is not supported.
    """
    field_type = None
    if input_type is str:
        field_type = "string"
    elif input_type is int:
        field_type = "integer"
    elif input_type is float:
        field_type = "number"
    elif input_type is bool:
        field_type = "boolean"

    return field_type


def extract_fields_from_class(
    dataclass: Type,
) -> Dict[str, Any]:
    """Extract arguments from dataclass.

    Args:
        dataclass: Dataclass to contains the arguments.

    Returns:
        Dictionary of the existing arguments including their type, description and default value.
    """

    # assign type and description
    arg_fields = {
        field.name: {"type": field.type, "description": field.metadata["help"]}
        for field in fields(dataclass)
    }

    # assign default values
    for field in fields(dataclass):

        if not isinstance(field.default, _MISSING_TYPE):

            if field.default is None:
                field.default = "none"

            arg_fields[field.name]["default"] = field.default

    # convert type to str
    for field_name in arg_fields:

        field_type = find_type(arg_fields[field_name]["type"])

        if field_type:

            arg_fields[field_name]["type"] = field_type

        elif (
            hasattr(arg_fields[field_name]["type"], "__origin__")
            and arg_fields[field_name]["type"].__origin__ is Union
        ):

            types = [
                find_type(type) for type in arg_fields[field_name]["type"].__args__
            ]
            types = [type for type in types if type is not None]

            if len(types) == 1:
                arg_fields[field_name]["type"] = types[0]
            else:
                raise ValueError(f"{arg_fields[field_name]['type']} not supported")

        else:
            raise ValueError(
                f" argument {field_name}: {arg_fields[field_name]['type']} not supported"
            )

    return arg_fields



"""PaccMann vanilla generator trained on polymer building blocks (catalysts/monomers)."""

import logging
from dataclasses import field
from typing import ClassVar, Dict, Optional, TypeVar

from ....domains.materials import SmallMolecule, validate_molecules
from ....exceptions import InvalidItem
from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
from ...registry import ApplicationsRegistry
from .implementation import Generator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = type(None)
S = TypeVar("S", bound=SmallMolecule)


class PolymerBlocks(GeneratorAlgorithm[S, T]):
    def __init__(
        self, configuration: AlgorithmConfiguration, target: Optional[T] = None
    ):
        """Polymer blocks generation.

        Args:
            configuration: domain and application
                specification, defining types and validations.
            target: unused since it is not a conditional generator.

        Example:
            An example for generating small molecules (SMILES) that resembles
            monomers/catalysts for polymer synthesis::

                configuration = PolymerBlocksGenerator()
                polymer_blocks = PolymerBlocks(configuration=configuration)
                items = list(polymer_blocks.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,
            target=None,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to sample batches via the Generator.

        Args:
            configuration: helps to set up the application.
            target: context or condition for the generation. Unused in the algorithm.

        Returns:
            callable generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.sample

    def validate_configuration(
        self, configuration: AlgorithmConfiguration
    ) -> AlgorithmConfiguration:
        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(PolymerBlocks)
class PolymerBlocksGenerator(AlgorithmConfiguration[SmallMolecule, None]):
    """Configuration to generate subunits of polymers."""

    algorithm_type: ClassVar[str] = "generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=32,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    generated_length: int = field(
        default=100,
        metadata=dict(
            description="Maximum length in tokens of the generated molcules (relates to the SMILES length)."
        ),
    )

    def get_target_description(self) -> Optional[Dict[str, str]]:
        """Get description of the target for generation.

        Returns:
            target description, returns None in case no target is used.
        """
        return None

    def get_conditional_generator(self, resources_path: str) -> Generator:
        return Generator(
            resources_path=resources_path,
            generated_length=self.generated_length,
            batch_size=self.batch_size,
        )

    def validate_item(self, item: str) -> SmallMolecule:
        (
            molecules,
            _,
        ) = validate_molecules([item])
        if molecules[0] is None:
            raise InvalidItem(
                title="InvalidSMILES",
                detail=f'rdkit.Chem.MolFromSmiles returned None for "{item}"',
            )
        return SmallMolecule(item)



"""Exceptions tests."""

import pytest

from gt4sd.exceptions import InvalidAlgorithmConfiguration, InvalidItem, S3SyncError


def test_s3_sync_error():
    error = S3SyncError("GenericSyncError", "my message")
    assert error.type == "S3SyncError"
    assert error.title == "GenericSyncError"
    assert error.detail == "my message"
    with pytest.raises(RuntimeError):
        str(error) == error.detail
        raise error


def test_invalid_item():
    error = InvalidItem("GenericInvaliItemError", "my message")
    assert error.type == "InvalidItem"
    assert error.title == "GenericInvaliItemError"
    assert error.detail == "my message"
    with pytest.raises(ValueError):
        str(error) == error.detail
        raise error


def test_invalid_algorithm_configuration():
    error = InvalidAlgorithmConfiguration(
        "GenericAlgorithmConfigurationError", "my message"
    )
    assert error.type == "InvalidAlgorithmConfiguration"
    assert error.title == "GenericAlgorithmConfigurationError"
    assert error.detail == "my message"
    with pytest.raises(ValueError):
        str(error) == error.detail
        raise error



"""HuggingFace generation algorithm."""

import logging
from dataclasses import field
from typing import ClassVar, Dict, Optional, Set, TypeVar

from ...core import (
    AlgorithmConfiguration,
    GeneratorAlgorithm,
    Untargeted,
    get_configuration_class_with_attributes,
)
from ...registry import ApplicationsRegistry
from .implementation import MODEL_TYPES, Generator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = type(None)
S = TypeVar("S", bound=str)


class HuggingFaceGenerationAlgorithm(GeneratorAlgorithm[S, T]):
    def __init__(
        self, configuration: AlgorithmConfiguration, target: Optional[T] = None
    ):
        """HuggingFace generation algorithm.

        Args:
            configuration: domain and application
                specification, defining types and validations.
            target: unused since it is not a conditional generator.

        Example:
            An example for using a generative algorithm from HuggingFace::

                configuration = HuggingFaceXLMGenerator()
                algorithm = HuggingFaceGenerationAlgorithm(configuration=configuration)
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,
            target=None,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to sample batches.

        Args:
            configuration: helps to set up the application.
            target: context or condition for the generation. Unused in the algorithm.

        Returns:
            callable generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.sample

    def validate_configuration(
        self, configuration: AlgorithmConfiguration
    ) -> AlgorithmConfiguration:
        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(HuggingFaceGenerationAlgorithm)
class HuggingFaceConfiguration(AlgorithmConfiguration[str, None]):
    """Basic configuration for an hugging face algorithm."""

    algorithm_type: ClassVar[str] = "generation"
    domain: ClassVar[str] = "nlp"

    model_type: str = field(
        default="",
        metadata=dict(
            description=f"Type of the model. Supported: {', '.join(MODEL_TYPES.keys())}"
        ),
    )
    prompt: str = field(
        default="I'm a stochastic parrot.",
        metadata=dict(description="Prompt for text generation."),
    )
    length: int = field(
        default=20, metadata=dict(description="Length of the generated text.")
    )
    stop_token: str = field(
        default="", metadata=dict(description="Stop token for text generation.")
    )
    temperature: float = field(
        default=1.0,
        metadata=dict(
            description="Temperature for sampling, the lower the greedier the sampling."
        ),
    )
    repetition_penalty: float = field(
        default=1.0,
        metadata=dict(
            description="Primarily useful for CTRL model, where 1.2 should be used."
        ),
    )
    k: int = field(
        default=50,
        metadata=dict(description="Number of top-k probability tokens to keep."),
    )
    p: float = field(
        default=1.0,
        metadata=dict(
            description="Only tokens with cumulative probabilities summing up to this value are kept."
        ),
    )
    prefix: str = field(
        default="",
        metadata=dict(
            description="Text defining context provided prior to the prompt."
        ),
    )
    number_of_sequences: int = field(
        default=8,
        metadata=dict(description="Number of text sequences to generate."),
    )

    def get_target_description(self) -> Optional[Dict[str, str]]:
        """Get description of the target for generation.

        Returns:
            target description, returns None in case no target is used.
        """
        return None

    def get_conditional_generator(self, resources_path: str, **kwargs) -> Generator:
        return Generator(
            resources_path=resources_path,
            model_type=self.model_type,
            model_name=self.algorithm_version,
            prompt=self.prompt,
            length=self.length,
            stop_token=self.stop_token,
            temperature=self.temperature,
            repetition_penalty=self.repetition_penalty,
            k=self.k,
            p=self.p,
            prefix=self.prefix,
            number_of_sequences=self.number_of_sequences,
        )


@ApplicationsRegistry.register_algorithm_application(HuggingFaceGenerationAlgorithm)
class HuggingFaceXLMGenerator(HuggingFaceConfiguration):
    """Configuration to generate text using XLM."""

    algorithm_version: str = "xlm-mlm-en-2048"
    model_type: str = "xlm"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://huggingface.co/models"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(HuggingFaceGenerationAlgorithm)
class HuggingFaceCTRLGenerator(HuggingFaceConfiguration):
    """Configuration to generate text using CTRL."""

    algorithm_version: str = "ctrl"
    model_type: str = "ctrl"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://huggingface.co/models"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(HuggingFaceGenerationAlgorithm)
class HuggingFaceGPT2Generator(HuggingFaceConfiguration):
    """Configuration to generate text using GPT2."""

    algorithm_version: str = "gpt2"
    model_type: str = "gpt2"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://huggingface.co/models"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(HuggingFaceGenerationAlgorithm)
class HuggingFaceOpenAIGPTGenerator(HuggingFaceConfiguration):
    """Configuration to generate text using OpenAIGPT."""

    algorithm_version: str = "openai-gpt"
    model_type: str = "openai-gpt"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://huggingface.co/models"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(HuggingFaceGenerationAlgorithm)
class HuggingFaceXLNetGenerator(HuggingFaceConfiguration):
    """Configuration to generate text using XLNet."""

    algorithm_version: str = "xlnet-large-cased"
    model_type: str = "xlnet"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://huggingface.co/models"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )


@ApplicationsRegistry.register_algorithm_application(HuggingFaceGenerationAlgorithm)
class HuggingFaceTransfoXLGenerator(HuggingFaceConfiguration):
    """Configuration to generate text using TransfoXL."""

    algorithm_version: str = "transfo-xl-wt103"
    model_type: str = "transfo-xl"

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://huggingface.co/models"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )



"""Recurrent Neural Networks with Proximal Policy Optimization algorithm implementation."""

from guacamol_baselines.smiles_lstm_ppo.goal_directed_generation import (
    PPODirectedGenerator,
)


class SMILESLSTMPPO:
    def __init__(
        self,
        model_path: str,
        num_epochs: int,
        episode_size: int,
        optimize_batch_size: int,
        entropy_weight: int,
        kl_div_weight: int,
        clip_param: float,
    ):
        """Initialize SMILESLSTMPPO.

        Args:
            model_path: path to load the model,
            num_epochs: number of epochs to sample
            episode_size: number of molecules sampled by the policy at the start of a series of ppo updates
            optimize_batch_size: batch size for the optimization
            entropy_weight: used for calculating entropy loss
            kl_div_weight: used for calculating Kullback-Leibler divergence loss
            clip_param: used for determining how far the new policy is from the old one
        """
        self.model_path = model_path
        self.num_epochs = num_epochs
        self.episode_size = episode_size
        self.optimize_batch_size = optimize_batch_size
        self.entropy_weight = entropy_weight
        self.kl_div_weight = kl_div_weight
        self.clip_param = clip_param

    def get_generator(self) -> PPODirectedGenerator:
        """
        used for creating an instance of the PPODirectedGenerator

        Returns:
            An instance of PPODirectedGenerator
        """
        optimiser = PPODirectedGenerator(
            pretrained_model_path=None,
            num_epochs=self.num_epochs,
            episode_size=self.episode_size,
            batch_size=self.optimize_batch_size,
            entropy_weight=self.entropy_weight,
            kl_div_weight=self.kl_div_weight,
            clip_param=self.clip_param,
        )
        return optimiser



"""Granular training utilities."""

import json
import logging
from argparse import Namespace
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Tuple

from pytorch_lightning import LightningDataModule, LightningModule

from ....frameworks.granular.dataloader.data_module import GranularDataModule
from ....frameworks.granular.dataloader.dataset import build_dataset_and_architecture
from ....frameworks.granular.ml.models import AUTOENCODER_ARCHITECTURES
from ....frameworks.granular.ml.module import GranularModule
from ...core import TrainingPipelineArguments
from ..core import PyTorchLightningTrainingPipeline

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GranularTrainingPipeline(PyTorchLightningTrainingPipeline):
    """Granular training pipelines."""

    def get_data_and_model_modules(
        self,
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
        **kwargs,
    ) -> Tuple[LightningDataModule, LightningModule]:
        """Get data and model modules for training.

        Args:
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Returns:
            the data and model modules.
        """

        configuration = {**model_args, **dataset_args}

        with open(model_args["model_list_path"], "r") as fp:  # type:ignore
            configuration["model_list"] = json.load(fp)["models"]

        arguments = Namespace(**configuration)
        datasets = []
        architecture_autoencoders = []
        architecture_latent_models = []
        for model in arguments.model_list:
            logger.info(f"dataset preparation for model={model}")
            hparams = configuration["model_list"][model]
            hparams["name"] = model
            model_type = hparams["type"].lower()
            dataset, architecture = build_dataset_and_architecture(
                hparams["name"],
                hparams["data_path"],
                hparams["data_file"],
                hparams["dataset_type"],
                hparams["type"],
                hparams,
            )
            datasets.append(dataset)
            if model_type in AUTOENCODER_ARCHITECTURES:
                architecture_autoencoders.append(architecture)
            else:
                architecture_latent_models.append(architecture)
        dm = GranularDataModule(
            datasets,
            batch_size=arguments.batch_size,
            validation_split=arguments.validation_split,
            validation_indices_file=arguments.validation_indices_file,
            stratified_batch_file=arguments.stratified_batch_file,
            stratified_value_name=arguments.stratified_value_name,
            num_workers=arguments.num_workers,
        )
        dm.prepare_data()
        module = GranularModule(
            architecture_autoencoders=architecture_autoencoders,
            architecture_latent_models=architecture_latent_models,
            lr=arguments.lr,
            test_output_path=arguments.test_output_path,
        )

        return dm, module


@dataclass
class GranularModelArguments(TrainingPipelineArguments):
    """
    Arguments related to model.
    """

    __name__ = "model_args"

    model_list_path: str = field(
        metadata={
            "help": "Path to a json file that contains a dictionary with models and their parameters."
        },
    )
    lr: float = field(
        default=0.0001,
        metadata={"help": "The learning rate."},
    )

    test_output_path: Optional[str] = field(
        default="./test",
        metadata={
            "help": "Path where to save latent encodings and predictions for the test set when an epoch ends. Defaults to a a folder called 'test' in the current working directory."
        },
    )


@dataclass
class GranularDataArguments(TrainingPipelineArguments):
    """
    Arguments related to data.
    """

    __name__ = "dataset_args"

    batch_size: int = field(
        default=64,
        metadata={"help": "Batch size of the training. Defaults to 64."},
    )
    validation_split: Optional[float] = field(
        default=None,
        metadata={
            "help": "Proportion used for validation. Defaults to None, a.k.a., use indices file if provided otherwise uses half of the data for validation."
        },
    )
    validation_indices_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "Indices to use for validation. Defaults to None, a.k.a., use validation split proportion, if not provided uses half of the data for validation."
        },
    )
    stratified_batch_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "Stratified batch file for sampling. Defaults to None, a.k.a., no stratified sampling."
        },
    )
    stratified_value_name: Optional[str] = field(
        default=None,
        metadata={
            "help": "Stratified value name. Defaults to None, a.k.a., no stratified sampling. Needed in case a stratified batch file is provided."
        },
    )
    num_workers: int = field(
        default=1,
        metadata={"help": "number of workers. Defaults to 1."},
    )



"""MolecularAI Implementation of sample generation, randomizing scaffolds as well as fetching unique sample sequences

The source of this file is
https://raw.githubusercontent.com/MolecularAI/Reinvent/982b26dd6cfeb8aa84b6d7e4a8c2a7edde2bad36/running_modes/lib_invent/rl_actions/sample_model.py
and it was only minimally changed. See README.md.
"""

__copyright__ = "Copyright 2021, MolecularAI"
__license__ = "Apache 2.0"

import logging
from dataclasses import dataclass
from typing import List

import numpy as np
import torch.utils.data as tud
from reinvent_chemistry import Conversions
from reinvent_chemistry.library_design import AttachmentPoints, BondMaker
from reinvent_chemistry.utils import get_indices_of_unique_smiles
from reinvent_models.lib_invent.models import dataset as md

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@dataclass
class SampledSequencesDTO:
    scaffold: str
    decoration: str
    nll: float


class ReinventBase:
    def __init__(
        self, model, batch_size: int, logger=None, randomize=False, sample_uniquely=True
    ):
        """
        Creates an instance of SampleModel.
        :params model: A model instance (better in scaffold_decorating mode).
        :params batch_size: Batch size to use.
        :return:
        """
        self.model = model
        self._batch_size = batch_size
        self._bond_maker = BondMaker()
        self._attachment_points = AttachmentPoints()
        self._randomize = randomize
        self._conversions = Conversions()
        self._sample_uniquely = sample_uniquely

    def get_dataloader(self, scaffold_list: List[str]) -> tud.DataLoader:
        """
        Get a dataloader for the list of scaffolds to use with reinvent.
        NOTE: This method was factored out of the `run` method from the original source.
        :params scaffold_list: A list of scaffold SMILES.
        :return: An instance of a torch dataloader.
        """
        scaffold_list = (
            self._randomize_scaffolds(scaffold_list)
            if self._randomize
            else scaffold_list
        )
        clean_scaffolds = [
            self._attachment_points.remove_attachment_point_numbers(scaffold)
            for scaffold in scaffold_list
        ]
        dataset = md.Dataset(
            clean_scaffolds,
            self.model.vocabulary.scaffold_vocabulary,
            self.model.vocabulary.scaffold_tokenizer,
        )
        dataloader = tud.DataLoader(
            dataset,
            batch_size=len(dataset),
            shuffle=False,
            collate_fn=md.Dataset.collate_fn,
        )
        return dataloader

    def run(self, scaffold_list: List[str]) -> List[SampledSequencesDTO]:
        """
        Samples the model for the given number of SMILES.
        NOTE: this method was slightly adapted from the original source.
        :params scaffold_list: A list of scaffold SMILES.
        :return: A list of SampledSequencesDTO.
        """

        dataloader = self.get_dataloader(scaffold_list)

        sampled_sequences = []
        for batch in dataloader:

            for _ in range(self._batch_size):
                scaffold_seqs, scaffold_seq_lengths = batch
                packed = self.model.sample_decorations(
                    scaffold_seqs, scaffold_seq_lengths
                )
                for scaffold, decoration, nll in packed:
                    sampled_sequences.append(
                        SampledSequencesDTO(scaffold, decoration, nll)
                    )

            if self._sample_uniquely:
                sampled_sequences = self._sample_unique_sequences(sampled_sequences)

        return sampled_sequences

    def _sample_unique_sequences(
        self, sampled_sequences: List[SampledSequencesDTO]
    ) -> List[SampledSequencesDTO]:
        strings = [
            "".join([ss.scaffold, ss.decoration])
            for index, ss in enumerate(sampled_sequences)
        ]
        unique_idxs = get_indices_of_unique_smiles(strings)
        sampled_sequences_np = np.array(sampled_sequences)
        unique_sampled_sequences = sampled_sequences_np[unique_idxs]
        return unique_sampled_sequences.tolist()

    def _randomize_scaffolds(self, scaffolds: List[str]):
        scaffold_mols = [
            self._conversions.smile_to_mol(scaffold) for scaffold in scaffolds
        ]
        randomized = [self._bond_maker.randomize_scaffold(mol) for mol in scaffold_mols]
        return randomized
